package sensor

import (
	"context"
	"fmt"
	"io/ioutil"
	"path/filepath"
	"strconv"
	"strings"
	"sync"

	"github.com/go-flucky/flucky/pkg/internal/format"
	"github.com/go-flucky/flucky/pkg/types"
	uuid "github.com/satori/go.uuid"
)

// DS18B20 is a sensor to measure humidity and temperature.
type DS18B20 struct {
	*types.Sensor
}

// GetSensorModel returns the sensor model
func (s *DS18B20) GetSensorModel() types.SensorModel {
	return s.Sensor.SensorModel
}

// Read measured values
func (s *DS18B20) Read() ([]*types.MeasuredValue, error) {

	if s.WireID == nil {
		return nil, fmt.Errorf("WireID is not specified for sensor %v", s.Name())
	}

	data, err := ioutil.ReadFile(filepath.Join("/sys/bus/w1/devices", *s.WireID, "/w1_slave"))
	if err != nil {
		return nil, fmt.Errorf("Can not read data from sensor %v", s.SensorName)
	}

	raw := string(data)

	i := strings.LastIndex(raw, "t=")
	if i == -1 {
		return nil, ErrReadSensor
	}

	c, err := strconv.ParseFloat(raw[i+2:len(raw)-1], 64)
	if err != nil {
		return nil, ErrParseData
	}

	temperatureValue := c / 1000

	measuredValues := []*types.MeasuredValue{
		&types.MeasuredValue{
			ID:        uuid.NewV4().String(),
			Value:     float64(temperatureValue),
			ValueType: types.MeasuredValueTypeTemperature,
			FromDate:  format.FormatedTime(),
			TillDate:  format.FormatedTime(),
			SensorID:  s.SensorID,
		},
	}

	return measuredValues, nil

}

// ReadChannel reads the measured values from the sensor and writes them to a
// channel.
func (s *DS18B20) ReadChannel(measuredValuesChannel chan<- []*types.MeasuredValue, errorChannel chan<- error, wg *sync.WaitGroup) {
	if wg != nil {
		defer wg.Done()
	}

	measuredValues, err := s.Read()
	if err != nil {
		errorChannel <- err
		return
	}

	measuredValuesChannel <- measuredValues

}

// ReadContinously reads the measured values continously from the sensor and
// writes them to a channel.
func (s *DS18B20) ReadContinously(ctx context.Context, measuredValuesChannel chan<- []*types.MeasuredValue, errorChannel chan<- error) {
	for {
		select {
		case <-ctx.Done():
			errorChannel <- fmt.Errorf("%v: Context closed: %v", s.SensorName, ctx.Err())
			return
		default:
			s.ReadChannel(measuredValuesChannel, errorChannel, nil)
		}
	}
}
